package com.swarmcast.fec.io;

import org.apache.log4j.Category;
import java.io.*;

/**
 * This class wraps a RandomAccessFile with synchronization.
 *
 * @Justin F. Chapweske.
 */
public class RAF {

    static Category cat = Category.getInstance(RAF.class.getName());

    File f;
    String mode;
    RandomAccessFile raf;
    boolean closed;

    public RAF(File f, String mode) throws IOException {
        this.f = f;
        this.mode = mode;
        this.raf = new RandomAccessFile(f,mode);
    }

    public synchronized void seekAndWrite(long pos, byte[] b, int off, 
                                          int len) throws IOException {
        raf.seek(pos);
        raf.write(b,off,len);
    }

    public synchronized void seekAndReadFully(long pos, byte[] b, int off,
                                              int len) throws IOException {
        raf.seek(pos);
        raf.readFully(b,off,len);
    }

    /**
     * This version of renameTo() attempts to mimic the behavior of the unix 
     * 'mv' command rather than File.renameTo.  This means that the destFile
     * will automatically be deleted if it exists and if we are unable to
     * mv the file directly because they are on different file systems then 
     * we will manually copy and delete the original.  If there is an
     * exception during the copy then we will attempt to delete the new copy
     * and leave the old copy intact.
     */
    public synchronized void renameTo(File destFile) throws IOException {
        if (closed) {
            throw new IOException("File closed.");
        }
        raf.close();
        // Move to final location.
        try {
            // If they are the same file than do nothing.  It is obviously 
            // important for this to be checked before we delete the destFile.
            if (f.getCanonicalFile().equals(destFile.getCanonicalFile())) {
                return;
            }

            // Delete the destFile if it exists.
            if (destFile.exists()) {
                cat.debug("renameTo(): destFile exists, deleting...");
                if (!destFile.delete()) {
                    throw new IOException("Unable to delete destination :"+
                                          destFile);
                }
            }
            
            if (!f.renameTo(destFile)) {
                cat.debug("renameTo(): File.renameTo failed, copying...");
                try {
                    byte[] b = new byte[8192];
                    InputStream is = new FileInputStream(f);
                    OutputStream os = new FileOutputStream(destFile);
                    int c;
                    while ((c = is.read(b,0,b.length)) != -1) {
                        os.write(b,0,c);
                    }
                    is.close();
                    os.close();
                } catch (IOException e) {
                    if (!destFile.delete()) {
                        throw new IOException("Unable to delete destination "+
                                              "after failed move :"+destFile+
                                              " :"+e.getMessage());
                    }       
                    throw new IOException("Unable to move "+f+" to dest: "+
                                          destFile+" :"+e.getMessage());
                }
                // If there was a prob with the copy then it should have bombed
                // before now.
                if (!f.delete()) {
                    throw new IOException("Unable to delete source post-move");
                }

                f = destFile;
            } else {
                f = destFile;
            }
        } finally {
            // If exception is thrown, re-open the old one.
            raf = new RandomAccessFile(f,mode);
        }
    }

    public synchronized String getMode() {
        return mode;
    }

    public synchronized void setReadOnly() throws IOException {
        if (closed) {
            throw new IOException("File closed.");
        }
        this.mode = "r";
        raf.close();
        raf = new RandomAccessFile(f,mode);
    }

    public synchronized void setLength(long len) throws IOException {
        raf.setLength(len);
    }

    public synchronized void close() throws IOException {
        closed = true;
        raf.close();
    }
}

        

